#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include "lcd1602.h"
#include "ADS1115.h"
#include "adc.h"
#include "utils.h"

#define BLMAX 253
#define BLMIN 1
// Для F_CPU = 16 МГц: 256 * 1024 / 16000000 = 0.016384 секунды (16.384 мс)
#define DISPTIME 610 // 10 sec

void init_timers();
int correction();
void blmax();
void blmin();
void start_timer2();
void stop_timer2();

volatile uint16_t overflow_count = 0;

int main(void) {
    char str[17] = {0};
    int32_t data;
    int corr = 0;
    int bl = 0;

    // Creating an liquid crystal display object
    struct lcd_t lcd = lcd_initialize(PD1, PD0, PC3, PC2, PC1, PC0);
/*             LCD1602                RS   EN   D4   D5   D6   D7    */
    lcd_clear(&lcd);
    init_timers();
    I2C_Init();
    adc_init();

    // зарядка
    DDRB &= ~(1 << DDB0);  // Пин 12 (PORTB0) как вход (заряжено)
    PORTB |= (1 << PORTB0); // Включение подтягивающего резистора
    DDRD &= ~(1 << DDD4);  // Пин 2 (PORTD4) как вход (идет зарядкаа)
    PORTD |= (1 << PORTD4); // Включение подтягивающего резистора

    // подсветка
    DDRD |= (1 << DDD5);  // Пин 9 (PORTD5) как выход (шим подсветки)
    OCR0B = 0;	// Duty cycle of 75% - 191
    TCCR0A |= (1<<COM0B1) | (1<<WGM01) | (1<<WGM00);	//Non-Inverting Fast PWM mode 3 using OCR B unit
    TCCR0B |= (1<<CS00);	//No-Prescalar

    // Проверяем флаг сброса от Watchdog Timer
    if (MCUSR & (1 << WDRF)) {
        // Если флаг WDRF установлен, это сброс от Watchdog Timer
        // Пропускаем выполнение функции
        MCUSR &= ~(1 << WDRF); // Сбрасываем флаг
    } else {
        // Если флаг WDRF не установлен, это подача питания
        corr = correction();
    }
    // Настраиваем Watchdog Timer для сброса через 2 секунды
    wdt_enable(WDTO_8S);

    lcd_set_cursor(&lcd, 0, 0);
    lcd_print(&lcd,"ESR: ");
    blmax();
    while (1) {
	lcd_set_cursor(&lcd, 0, 5);
	data = ads1115_readADC_Diff_A1_3(ADS1115_ADDR_GND, ADS1115_DR_32SPS, ADS1115_PGA_0_256);
	if (data < 32767) {
	    sprintf(str, "%2.3f Ohm  ", (float)(data - corr) * 0.001);
	    lcd_print(&lcd, str);
	    if (bl == 1) {
		if (OCR0B == BLMIN+1) blmax(); else
		stop_timer2();
		bl = 0;
	    }
	} else {
	    lcd_print(&lcd, "READY           ");
	    if (bl == 0) {
		if (OCR0B == BLMAX-1) start_timer2();
		bl = 1;
	    }
	}
	lcd_set_cursor(&lcd, 1, 0);
	if (!(PIND & (1 << PIND4))) {
	    uint8_t percents = map(adc_read(6), 486,789, 0,100);
	    sprintf(str, "CHARGING:%3d %%", percents);
	    lcd_print(&lcd, str);
	} else if (!(PINB & (1 << PINB0))) {
	    lcd_print(&lcd, "FULLY CHARGED  ");
	} else {
	    uint8_t percents = map(adc_read(6), 486,789, 0,100);
	    sprintf(str, "BAT:%3d %%     ", percents);
	    lcd_print(&lcd, str);
	}
        // Сбрасываем Watchdog Timer
        wdt_reset();
    }
    return 0;
}

void blmax () {
    for (int bl = BLMIN; bl != BLMAX; bl++) {
        OCR0B = bl;
        _delay_ms(.2);
    }
}
void blmin () {
    for (int bl = BLMAX; bl != BLMIN; bl--) {
        OCR0B = bl;
        _delay_ms(1);
    }
}

#define CORR 2
int correction() {
    int32_t data = 0;
    int32_t data_tmp = 100;
    int corr = 0, scorr;
    while (data_tmp != data && (data_tmp >=60 || data_tmp  <= -60)) {
	data = ads1115_readADC_Diff_A1_3(ADS1115_ADDR_GND, ADS1115_DR_8SPS, ADS1115_PGA_0_256);
	_delay_ms(200);
	data_tmp = ads1115_readADC_Diff_A1_3(ADS1115_ADDR_GND, ADS1115_DR_8SPS, ADS1115_PGA_0_256);
    }
    data = ads1115_readADC_Diff_A1_3(ADS1115_ADDR_GND, ADS1115_DR_32SPS, ADS1115_PGA_0_256);
    scorr = eeprom_read_word((uint16_t*)10);
    corr = scorr;
    if (data >= 0 && data != 32767) {
	corr = data + CORR;
	if (scorr != corr) eeprom_update_word((uint16_t*)10, corr);
    }
    if (data <= 0 && data != -32767) {
	corr = data - CORR + 1;
	if (scorr != corr) eeprom_write_word((uint16_t*)10, corr);
    }
    return corr;
}

void init_timers() {
    // Настройка пинов на выход
    DDRB |= (1 << DDB1);  // Пин 13 (PORTB1) как выход
    DDRD |= (1 << DDD7);  // Пин 11 (PORTD7) как выход
    DDRD |= (1 << DDD6);  // Пин 10 (PORTD6) как выход

    DDRD &= ~(1 << DDD3);  // Пин 1 (PORTD3) (INT1) как вход


    // Настраиваем Timer2 на нормальный режим (Normal Mode) шим посветки
    TCCR2A = 0x00;  // Обычный режим
    TCCR2B = 0x00;

    // Устанавливаем предделитель (Prescaler) на 1024
    TCCR2B |= (1 << CS22) | (1 << CS21) | (1 << CS20);
/*
    // Сбрасываем счетчик
    TCNT2 = 0;

    // Настраиваем прерывание по переполнению
    TIMSK2 |= (1 << TOIE2);
*/

    // Настройка Timer1 для генерации меандра 100 кГц на пине 13 (OC1A)
    TCCR1A = 0x00;  // Обычный режим
    TCCR1B = 0x00;
    TCNT1 = 0;   // Сброс счетчика таймера

    // Режим CTC (Clear Timer on Compare Match)
    TCCR1B |= (1 << WGM12);

    // Установка предделителя на 1 (без деления)
    TCCR1B |= (1 << CS10);

    // Установка значения сравнения для частоты 100 кГц
    // Частота таймера = 16 МГц / (2 * (1 + OCR1A))
    // Для 100 кГц: OCR1A = (16 МГц / (2 * 100 кГц)) - 1 = 79
    OCR1A = 79;

    // Настройка вывода OC1A (пин 9) на режим "переключение при совпадении"
    TCCR1A |= (1 << COM1A0);  // Переключение состояния пина 9 при совпадении с OCR1A

    // Настройка внешних прерываний для детектирования фронтов меандра
    EICRA |= (1 << ISC10);  // Прерывание INT1 по любому изменению на пине 3 (INT1)
    EIMSK |= (1 << INT1);    // Включение прерывания INT1

    // Включение глобальных прерываний
    sei();
}

// Обработчик прерывания INT1 (пин 1)
ISR(INT1_vect) {
    // Задержка перед генерацией импульса (середина фронта)
    __asm__(
        "nop\n\t"  // 62.5 нс
        "nop\n\t"  // 62.5 нс
        "nop\n\t"  // 62.5 нс
        "nop\n\t"  // 62.5 нс
        "nop\n\t"  // 62.5 нс
        "nop\n\t"  // 62.5 нс
        "nop\n\t"  // 62.5 нс
        "nop\n\t"  // 62.5 нс
//        "nop\n\t"  // 62.5 нс
//        "nop\n\t"  // 62.5 нс
    );  // Итого: 8 * 62.5 нс = 500 нс (0.5 мкс)

    // Генерация импульсов в середине фронтов
//    if (meanderState) {
    if (PINB & (1 << PINB1)) { // Определяем текущее состояние меандра на пине 13
        // Положительный фронт
        PORTD |= (1 << PORTD7);  // Установка HIGH на пине 9
        __asm__(
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
        );  // Итого: 16 * 62.5 нс = 1000 нс (1 мкс)
//_delay_us(1);
        PORTD &= ~(1 << PORTD7); // Установка LOW на пине 9
    } else {
        // Отрицательный фронт
        PORTD |= (1 << PORTD6);  // Установка HIGH на пине 10
        __asm__(
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
            "nop\n\t"  // 62.5 нс
        );  // Итого: 16 * 62.5 нс = 1000 нс (1 мкс)
//_delay_us(1);
        PORTD &= ~(1 << PORTD6); // Установка LOW на пине 10
    }
}

// Обработчик прерывания Timer2
ISR(TIMER2_OVF_vect) {
//    static uint16_t overflow_count = 0;

    // Каждое переполнение происходит каждые (256 * 1024) / F_CPU секунд
    // Для F_CPU = 16 МГц: 256 * 1024 / 16000000 = 0.016384 секунды (16.384 мс)
    overflow_count++;

    // 60 секунд / 0.016384 секунды = ~3662 переполнений
    if (overflow_count >= DISPTIME) { // 30 секунд 1831
	blmin();
//        overflow_count = 0;  // Сбрасываем счетчик
	stop_timer2();
    }
}

void stop_timer2() {
    // Обнуляем счетчик
    TCNT2 = 0;
    // Отключаем прерывание по переполнению
    TIMSK2 &= ~(1 << TOIE2);
    overflow_count = 0;  // Сбрасываем счетчик
}

void start_timer2() {
    // Сбрасываем счетчик
    TCNT2 = 0;
    // Настраиваем прерывание по переполнению
    TIMSK2 |= (1 << TOIE2);
    overflow_count = 0;  // Сбрасываем счетчик
}
